# pyfoam\pyfoam\core\conditions\condition.py

from abc import ABC, abstractmethod
from typing import Callable, Dict, Any
import numpy as np

class BoundaryCondition(ABC):
    """
    Abstract base class for numeric boundary conditions.
    
    Subclasses must implement the apply() method that takes a mesh and a time t,
    and returns a dictionary mapping node indices (int) to the prescribed boundary value.
    """
    @abstractmethod
    def apply(self, mesh: Any, t: float) -> Dict[int, float]:
        """
        Apply the boundary condition to the given mesh at time t.

        Parameters:
            mesh: A mesh object with attributes 'nodes' (a NumPy array) and 'bounds' (a list of (lower, upper) tuples).
            t (float): The current time.

        Returns:
            Dict[int, float]: A mapping from node indices to the prescribed value.
        """
        pass


class DirichletCondition(BoundaryCondition):
    """
    Numeric Dirichlet boundary condition.

    The user supplies a function that takes a node's coordinate (as a NumPy array) and a time t,
    and returns the prescribed value at that node.
    """
    def __init__(self, func: Callable[[np.ndarray, float], float]):
        """
        Initialize the Dirichlet condition.

        Parameters:
            func (Callable[[np.ndarray, float], float]): A function g(x, t) that returns the prescribed value.
        """
        self.func = func

    def apply(self, mesh: Any, t: float) -> Dict[int, float]:
        """
        Apply the Dirichlet boundary condition to the mesh.

        A node is considered to be on the boundary if its coordinate in any dimension is
        numerically close to the lower or upper bound (using a tolerance).

        Parameters:
            mesh: A mesh object with attributes:
                  - nodes: NumPy array of shape (N, dim).
                  - bounds: List of (lower, upper) tuples for each dimension.
            t (float): The current time.

        Returns:
            Dict[int, float]: A dictionary mapping node indices to the prescribed Dirichlet value.
        """
        bc_values = {}
        bounds = mesh.bounds  # Example: [(0.0, 1.0), (0.0, 1.0)] for a 2D mesh.
        for i, node in enumerate(mesh.nodes):
            on_boundary = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(node[d], lower, atol=1e-8) or np.isclose(node[d], upper, atol=1e-8):
                    on_boundary = True
                    break
            if on_boundary:
                bc_values[i] = self.func(node, t)
        return bc_values


class NeumannCondition(BoundaryCondition):
    """
    Numeric Neumann boundary condition.

    The user supplies a function that takes a node's coordinate (as a NumPy array) and a time t,
    and returns the prescribed flux value at that node.
    """
    def __init__(self, func: Callable[[np.ndarray, float], float]):
        """
        Initialize the Neumann condition.

        Parameters:
            func (Callable[[np.ndarray, float], float]): A function h(x, t) that returns the prescribed flux.
        """
        self.func = func

    def apply(self, mesh: Any, t: float) -> Dict[int, float]:
        """
        Apply the Neumann boundary condition to the mesh.

        Parameters:
            mesh: A mesh object with attributes 'nodes' and 'bounds'.
            t (float): The current time.

        Returns:
            Dict[int, float]: A dictionary mapping node indices to the prescribed Neumann flux.
        """
        bc_flux = {}
        bounds = mesh.bounds
        for i, node in enumerate(mesh.nodes):
            on_boundary = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(node[d], lower, atol=1e-8) or np.isclose(node[d], upper, atol=1e-8):
                    on_boundary = True
                    break
            if on_boundary:
                bc_flux[i] = self.func(node, t)
        return bc_flux

class InitialCondition(ABC):
    """
    Abstract base class for numeric initial conditions.
    
    Subclasses must implement the apply() method that takes a mesh (with a NumPy array of nodes)
    and returns a dictionary mapping node indices to the initial value.
    """
    @abstractmethod
    def apply(self, mesh: Any, t: float = 0) -> Dict[int, float]:
        """
        Apply the initial condition to the given mesh at time t.

        Parameters:
            mesh: A mesh object with attribute 'nodes' (a NumPy array).
            t (float): The time at which the initial condition is evaluated (default is 0).

        Returns:
            Dict[int, float]: A mapping from node indices to the initial value.
        """
        pass


class ScalarInitialCondition(InitialCondition):
    """
    Numeric scalar initial condition.

    The user supplies a function that takes a node's coordinate (as a NumPy array) and an optional time t,
    and returns the initial value at that node.
    """
    def __init__(self, func: Callable[[np.ndarray, float], float]):
        """
        Initialize the scalar initial condition.

        Parameters:
            func (Callable[[np.ndarray, float], float]): A function u0(x, t) that returns the initial value.
        """
        self.func = func

    def apply(self, mesh: Any, t: float = 0) -> Dict[int, float]:
        """
        Apply the initial condition to the mesh.

        Parameters:
            mesh: A mesh object with attribute 'nodes' (a NumPy array).
            t (float): Time at which the initial condition is evaluated (default is 0).

        Returns:
            Dict[int, float]: A dictionary mapping node indices to the initial value.
        """
        ic_values = {}
        for i, node in enumerate(mesh.nodes):
            ic_values[i] = self.func(node, t)
        return ic_values