# pyfoam\pyfoam\core\meshing\mesh.py

import numpy as np
from typing import List, Tuple
from itertools import product
from typing import Optional
from scipy.spatial import Delaunay

class StructuredMesh:
    
    def __init__(self, 
                 dimensions: int, 
                 bounds: List[Tuple[float, float]], 
                 num_cells: List[int]):

        if len(bounds) != dimensions:
            raise ValueError("bounds must have the same length as dimensions")

        if len(num_cells) != dimensions:
            raise ValueError("num_cells must have the same length as dimensions")

        self.nodes = None
        self.cells = None

        self.generate_mesh()


    def generate_mesh(self):

        self.generate_nodes()
        self.generate_cells()

    def generate_nodes(self):
        grids = []
        for d in range(self.dimensions):
            lower, upper = self.bounds[d]
            n = self.num_cells[d]
            # Generate n+1 points in each dimension.
            grids.append(np.linspace(lower, upper, n + 1))
        # Create meshgrid with "ij" indexing.
        mesh_grids = np.meshgrid(*grids, indexing='ij')
        # Stack into an array of shape (N, dimensions) where N is total number of nodes.
        self.nodes = np.stack([g.flatten() for g in mesh_grids], axis=-1)

    def generate_cells(self):
        self.cells = []
        # For each cell, indices vary from 0 to num_cells[d]-1.
        grid_ranges = [range(n) for n in self.num_cells]
        for idx in product(*grid_ranges):
            vertices = []
            # Generate all vertex offsets (0 or 1) for each dimension.
            for offsets in product([0, 1], repeat=self.dimensions):
                vertex_index = tuple(idx[d] + offsets[d] for d in range(self.dimensions))
                # Convert multi-index to linear index (row-major order).
                linear_index = 0
                multiplier = 1
                for d in reversed(range(self.dimensions)):
                    linear_index += vertex_index[d] * multiplier
                    multiplier *= (self.num_cells[d] + 1)
                vertices.append(linear_index)
            self.cells.append(vertices)
            
class UnstructuredMesh:
    def __init__(self, 
                 dimensions: int, 
                 num_points: int, 
                 bounds: List[Tuple[float, float]], 
                 seed: Optional[int] = None):
        
        if len(bounds) != dimensions:
            raise ValueError("bounds must have the same length as dimensions")
        
        self.dimensions = dimensions
        self.num_points = num_points
        self.bounds = bounds
        self.seed = seed
        
        self.nodes = None
        self.cells = None
        
        self.generate_mesh()
        
    def generate_mesh(self):
        if self.seed is not None:
            np.random.seed(self.seed)
        # For each dimension, sample num_points uniformly within bounds.
        # Here, we generate points in the entire domain independently.
        # (One could also generate a structured random grid; here we assume an amorphous distribution.)
        coords = []
        for d in range(self.dimension):
            lower, upper = self.bounds[d]
            # Generate num_points random values in [lower, upper].
            coord_d = np.random.uniform(lower, upper, self.num_points)
            coords.append(coord_d)
        # Stack the coordinate arrays into a (num_points, dimension) array.
        self.nodes = np.column_stack(coords)
        
    def generate_connectivity(self):
        tri = Delaunay(self.nodes)
        self.cells = tri.simplices.tolist()