# pyfoam\pyfoam\core\solvers\cfds\cfd_solver.py

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from navier_stocks_solver import NavierStokesFDM, NavierStokesFEM, NavierStokesFVM

class CFDSolver:
    def __init__(self,
                 method: str,
                 mesh: StructuredMesh,
                 solver_params: dict):
        """
        Initialize a generic CFD solver that can use FVM, FDM, or FEM for solving the steady
        Navier–Stokes equations (penalty formulation).

        Parameters:
            method (str): One of "FVM", "FDM", "FEM" (case-insensitive).
            mesh (StructuredMesh): The computational mesh.
            solver_params (dict): A dictionary containing the following keys:
                  - nu         : Kinematic viscosity.
                  - penalty    : Penalty parameter.
                  - f_u, f_v   : Source term functions for the momentum equations.
                  - dirichlet_bc_u, dirichlet_bc_v : DirichletCondition objects for u and v.
                  - (Optional) neumann_bc_u, neumann_bc_v : NeumannCondition objects.
        """
        self.method = method.upper()
        self.mesh = mesh
        self.params = solver_params
        self.solver = None

        if self.method == "FVM":
            self.solver = NavierStokesFVM(mesh,
                                          self.params['nu'],
                                          self.params['penalty'],
                                          self.params['f_u'],
                                          self.params['f_v'],
                                          self.params['dirichlet_bc_u'],
                                          self.params['dirichlet_bc_v'],
                                          self.params.get('neumann_bc_u', None),
                                          self.params.get('neumann_bc_v', None))
        elif self.method == "FDM":
            self.solver = NavierStokesFDM(mesh,
                                          self.params['nu'],
                                          self.params['penalty'],
                                          self.params['f_u'],
                                          self.params['f_v'],
                                          self.params['dirichlet_bc_u'],
                                          self.params['dirichlet_bc_v'],
                                          self.params.get('neumann_bc_u', None),
                                          self.params.get('neumann_bc_v', None))
        elif self.method == "FEM":
            self.solver = NavierStokesFEM(mesh,
                                          self.params['nu'],
                                          self.params['penalty'],
                                          self.params['f_u'],
                                          self.params['f_v'],
                                          self.params['dirichlet_bc_u'],
                                          self.params['dirichlet_bc_v'],
                                          self.params.get('neumann_bc_u', None),
                                          self.params.get('neumann_bc_v', None))
        else:
            raise ValueError("Unknown method: choose 'FVM', 'FDM', or 'FEM'.")

    def solve(self) -> np.ndarray:
        """
        Solve the CFD problem using the selected method.
        
        Returns:
            np.ndarray: The computed velocity field as a two-column array [u, v].
        """
        return self.solver.solve()
