# pyfoam\pyfoam\core\solvers\fem_helmholtz_solver.py

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

class HelmholtzFEMSolver:
    def __init__(self,
                 mesh: StructuredMesh,
                 f: Callable[[np.ndarray], float],
                 k: float,
                 dirichlet_bc: DirichletCondition,
                 neumann_bc: Optional[NeumannCondition] = None):
        """
        Initialize the Helmholtz FEM solver.

        Parameters:
            mesh (StructuredMesh): A structured mesh with attributes:
                                   - nodes: NumPy array of shape (N, dim) with node coordinates.
                                   - cells: List of cells, each cell is a list of node indices (triangular elements).
            f (Callable[[np.ndarray], float]): Source term function f(x,y).
            k (float): Wave number.
            dirichlet_bc (DirichletCondition): Dirichlet BC object prescribing u = g(x,y) on ∂Ω.
            neumann_bc (Optional[NeumannCondition]): Optional Neumann BC object prescribing flux h(x,y) on ∂Ω.
        """
        self.mesh = mesh
        self.f = f
        self.k = k
        self.dirichlet_bc = dirichlet_bc
        self.neumann_bc = neumann_bc

        self.num_nodes = self.mesh.nodes.shape[0]
        self.A = None  # Global matrix (stiffness and mass combined)
        self.M = None  # Global consistent mass matrix
        self.solution = None

    def assemble_system(self) -> None:
        """
        Assemble the global matrices for the Helmholtz equation.

        For each triangular cell, compute:
          - The area (using the determinant formula).
          - The gradients of the linear shape functions.
          - The local diffusion (stiffness) matrix:
                Ke_diff = area * (grad_N^T @ grad_N)
          - The local consistent mass matrix:
                Me = (area/12) * (ones(3,3) + I)
          - The local contribution to the global matrix is then:
                Ke = Ke_diff - k^2 * Me
          - The local load vector is computed by:
                fe = (area/3) * f(centroid) * [1,1,1]^T

        The local contributions are assembled into global matrices A and the load vector b.
        """
        rows_A, cols_A, data_A = [], [], []
        rows_M, cols_M, data_M = [], [], []
        b = np.zeros(self.num_nodes)

        for cell in self.mesh.cells:
            coords = self.mesh.nodes[cell]  # shape: (3,2)
            x = coords[:, 0]
            y = coords[:, 1]
            # Compute area using the determinant formula.
            area = 0.5 * np.abs(np.linalg.det(np.array([[1, x[0], y[0]],
                                                         [1, x[1], y[1]],
                                                         [1, x[2], y[2]]])))
            centroid = np.mean(coords, axis=0)
            # Evaluate source term at the centroid.
            f_val = self.f(centroid)
            # Compute gradients of shape functions.
            b_vec = np.array([y[1] - y[2], y[2] - y[0], y[0] - y[1]])
            c_vec = np.array([x[2] - x[1], x[0] - x[2], x[1] - x[0]])
            grad_N = np.vstack((b_vec, c_vec)) / (2 * area)  # shape: (2,3)
            # Local diffusion (stiffness) matrix.
            Ke_diff = area * (grad_N.T @ grad_N)
            # Local consistent mass matrix.
            Me = (area / 12) * (np.ones((3, 3)) + np.eye(3))
            # Local Helmholtz matrix: diffusion minus k^2 * mass.
            Ke = Ke_diff - (self.k**2) * Me

            # Local load vector.
            fe = (area / 3) * f_val * np.ones(3)

            for i_local, i_global in enumerate(cell):
                b[i_global] += fe[i_local]
                for j_local, j_global in enumerate(cell):
                    rows_A.append(i_global)
                    cols_A.append(j_global)
                    data_A.append(Ke[i_local, j_local])
                    # Also assemble mass matrix if needed (could be useful for other purposes).
                    rows_M.append(i_global)
                    cols_M.append(j_global)
                    data_M.append(Me[i_local, j_local])
        A = sp.coo_matrix((data_A, (rows_A, cols_A)), shape=(self.num_nodes, self.num_nodes))
        self.A = A.tocsr()
        self.M = sp.coo_matrix((data_M, (rows_M, cols_M)), shape=(self.num_nodes, self.num_nodes)).tocsr()
        self.b = b

    def apply_boundary_conditions(self) -> None:
        """
        Apply boundary conditions to the assembled system.

        The DirichletCondition object's apply() method returns a dictionary mapping node indices
        to prescribed values g(x,y). For each boundary node, the corresponding row in the global
        matrix A is zeroed (with the diagonal set to 1) and the load vector b is set to g(x,y).
        Additionally, if a NeumannCondition is provided, its contributions are added to b.
        """
        bc_dirichlet = self.dirichlet_bc.apply(self.mesh, t=0)
        for node_idx, value in bc_dirichlet.items():
            self.A.data[self.A.indptr[node_idx]:self.A.indptr[node_idx+1]] = 0
            self.A[node_idx, node_idx] = 1.0
            self.b[node_idx] = value
        if self.neumann_bc is not None:
            bc_neumann = self.neumann_bc.apply(self.mesh, t=0)
            for node_idx, flux in bc_neumann.items():
                self.b[node_idx] += flux

    def solve(self) -> np.ndarray:
        """
        Solve the stationary Helmholtz equation.

        The method assembles the system, applies boundary conditions, and then solves
        the linear system A u = b using a sparse direct solver.

        Returns:
            np.ndarray: The computed nodal solution vector.
        """
        self.assemble_system()
        self.apply_boundary_conditions()
        u = spla.spsolve(self.A, self.b)
        self.solution = u
        return u