# pyfoam\pyfoam\core\solvers\fem_solver.py

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

class FEMSolver:
    def __init__(self,
                 mesh: StructuredMesh,
                 a: Callable[[np.ndarray], float],
                 c: Callable[[np.ndarray], float],
                 f: Callable[[np.ndarray, float], float],
                 dirichlet_bc: DirichletCondition,
                 neumann_bc: Optional[NeumannCondition] = None,
                 u0: Optional[Callable[[np.ndarray], float]] = None,
                 time_dependent: bool = False):
        """
        Initialize the generic FEM solver.

        Parameters:
            mesh: A mesh object (e.g., from create_structured_mesh) with attributes:
                  - nodes: NumPy array of shape (N, dim) containing node coordinates.
                  - cells: List of cells (each cell is a list of node indices defining a triangular element).
            a (Callable[[np.ndarray], float]): Diffusion coefficient function a(x).
            c (Callable[[np.ndarray], float]): Reaction coefficient function c(x).
            f (Callable[[np.ndarray, float], float]): Source term function f(x, t).
            dirichlet_bc (DirichletCondition): Dirichlet boundary condition object.
            neumann_bc (Optional[NeumannCondition]): Optional Neumann boundary condition object.
            u0 (Optional[Callable[[np.ndarray], float]]): Initial condition function u₀(x) (required for time-dependent problems).
            time_dependent (bool): Flag indicating if the problem is time-dependent.
        """
        self.mesh = mesh
        self.a = a
        self.c = c
        self.f = f
        self.dirichlet_bc = dirichlet_bc
        self.neumann_bc = neumann_bc
        self.u0 = u0
        self.time_dependent = time_dependent

        self.num_nodes = self.mesh.nodes.shape[0]
        self.A = None  # Global stiffness matrix
        self.M = None  # Global mass matrix (used for time-dependent problems)
        self.solution = None

    def assemble_matrices(self) -> None:
        """
        Assemble the global stiffness matrix A and the consistent mass matrix M.

        For each triangular element in the mesh, the following local matrices are computed:
          - The area is computed using the determinant method.
          - The local stiffness matrix is computed as:
                Ke = area * (grad_N^T @ (a_avg * grad_N)) + (area/12)*c_avg*(ones + I)
            where a_avg = a(centroid) and c_avg = c(centroid), and grad_N are the gradients of the
            linear shape functions.
          - The consistent mass matrix is computed as:
                Me = (area/12)*(ones + I)
        The local matrices are assembled into global matrices A and M, stored in CSR format.
        """
        rows_A, cols_A, data_A = [], [], []
        rows_M, cols_M, data_M = [], [], []

        for cell in self.mesh.cells:
            # Extract the coordinates of the vertices in the cell.
            coords = self.mesh.nodes[cell]  # shape: (3, 2)
            x = coords[:, 0]
            y = coords[:, 1]
            # Compute the area of the triangle.
            area = 0.5 * np.abs(np.linalg.det(np.array([[1, x[0], y[0]],
                                                         [1, x[1], y[1]],
                                                         [1, x[2], y[2]]])))
            centroid = np.mean(coords, axis=0)
            a_avg = self.a(centroid)
            c_avg = self.c(centroid)
            # Compute gradients of the shape functions.
            b_vec = np.array([y[1] - y[2], y[2] - y[0], y[0] - y[1]])
            c_vec = np.array([x[2] - x[1], x[0] - x[2], x[1] - x[0]])
            grad_N = np.vstack((b_vec, c_vec)) / (2 * area)  # shape: (2, 3)
            # Local stiffness matrix: diffusion and reaction.
            Ke_diff = area * (grad_N.T @ (a_avg * grad_N))
            Ke_react = (area / 12) * c_avg * (np.ones((3, 3)) + np.eye(3))
            Ke = Ke_diff + Ke_react
            # Local consistent mass matrix.
            Me = (area / 12) * (np.ones((3, 3)) + np.eye(3))

            for i_local, i_global in enumerate(cell):
                for j_local, j_global in enumerate(cell):
                    rows_A.append(i_global)
                    cols_A.append(j_global)
                    data_A.append(Ke[i_local, j_local])
                    rows_M.append(i_global)
                    cols_M.append(j_global)
                    data_M.append(Me[i_local, j_local])
        A = sp.coo_matrix((data_A, (rows_A, cols_A)), shape=(self.num_nodes, self.num_nodes))
        M = sp.coo_matrix((data_M, (rows_M, cols_M)), shape=(self.num_nodes, self.num_nodes))
        self.A = A.tocsr()
        self.M = M.tocsr()

    def assemble_load_vector(self, t: float) -> np.ndarray:
        """
        Assemble the global load vector b at time t.

        For each cell, the local load vector is computed using centroid quadrature:
            fe = (area/3) * f(centroid, t) * [1, 1, 1]^T
        The contributions are then summed into the global load vector.

        Parameters:
            t (float): Time at which the source term is evaluated.

        Returns:
            np.ndarray: Global load vector b of length num_nodes.
        """
        b = np.zeros(self.num_nodes)
        for cell in self.mesh.cells:
            coords = self.mesh.nodes[cell]
            x = coords[:, 0]
            y = coords[:, 1]
            area = 0.5 * np.abs(np.linalg.det(np.array([[1, x[0], y[0]],
                                                         [1, x[1], y[1]],
                                                         [1, x[2], y[2]]])))
            centroid = np.mean(coords, axis=0)
            fe = (area / 3) * self.f(centroid, t) * np.ones(3)
            for i_local, i_global in enumerate(cell):
                b[i_global] += fe[i_local]
        return b

    def apply_boundary_conditions(self, b: np.ndarray, t: float) -> np.ndarray:
        """
        Apply boundary conditions to the system.

        The DirichletCondition object's apply() method returns a dictionary mapping node indices
        to prescribed values g_D(x, t). For each such node, the corresponding row in the stiffness
        matrix A is zeroed and its diagonal is set to 1, and the load vector b is set to g_D(x, t).
        Similarly, the NeumannCondition (if provided) returns a dictionary mapping node indices
        to flux contributions h(x, t) which are added to b.

        Parameters:
            b (np.ndarray): Global load vector.
            t (float): Current time.

        Returns:
            np.ndarray: Modified load vector b with boundary conditions enforced.
        """
        # Apply Dirichlet conditions.
        bc_dirichlet = self.dirichlet_bc.apply(self.mesh, t)
        for node_idx, value in bc_dirichlet.items():
            self.A.data[self.A.indptr[node_idx]:self.A.indptr[node_idx+1]] = 0
            self.A[node_idx, node_idx] = 1.0
            b[node_idx] = value
        # Apply Neumann conditions if available.
        if self.neumann_bc is not None:
            bc_neumann = self.neumann_bc.apply(self.mesh, t)
            for node_idx, flux in bc_neumann.items():
                b[node_idx] += flux
        return b

    def solve_stationary(self) -> np.ndarray:
        """
        Solve the stationary PDE:
            -∇·(a(x)∇u(x)) + c(x) u(x) = f(x)
        The load vector is assembled at t = 0, boundary conditions are applied, and the linear
        system A u = b is solved.

        Returns:
            np.ndarray: Computed nodal solution vector.
        """
        if self.A is None or self.M is None:
            self.assemble_matrices()
        b = self.assemble_load_vector(t=0)
        b = self.apply_boundary_conditions(b, t=0)
        u = spla.spsolve(self.A, b)
        self.solution = u
        return u

    def solve_time(self, dt: float, T: float) -> np.ndarray:
        """
        Solve the time-dependent PDE:
            u_t - ∇·(a(x)∇u) + c(x) u = f(x,t)
        using the implicit Euler method.

        The time-discrete system is:
            (M/dt + A) u^(n+1) = (M/dt)*u^(n) + b^(n+1)
        where b^(n+1) is the load vector (including source and Neumann contributions) at time t^(n+1).
        Dirichlet boundary conditions are enforced at every time step.

        Parameters:
            dt (float): Time step size.
            T (float): Final time.

        Returns:
            np.ndarray: A 2D array of shape (num_steps+1, num_nodes) containing the solution at each time step.
        """
        num_steps = int(T / dt)
        if self.A is None or self.M is None:
            self.assemble_matrices()

        # Preassemble constant part of LHS.
        LHS = self.M / dt + self.A
        LHS = LHS.tocsr()

        if self.u0 is None:
            raise ValueError("Initial condition u₀ must be provided for time-dependent problems.")
        u = np.array([self.u0(x) for x in self.mesh.nodes])
        u = self.apply_boundary_conditions(u, t=0)

        solution_time = np.zeros((num_steps + 1, self.num_nodes))
        solution_time[0, :] = u.copy()

        t = 0.0
        for n in range(num_steps):
            t_next = t + dt
            b_time = self.assemble_load_vector(t_next)
            b_time = self.apply_boundary_conditions(b_time, t_next)
            RHS = self.M.dot(u) / dt + b_time
            u_new = spla.spsolve(LHS, RHS)
            u = u_new.copy()
            solution_time[n + 1, :] = u
            t = t_next

        self.solution = u
        return solution_time

    def solve(self) -> np.ndarray:
        """
        Generic solve method.

        For stationary problems (time_dependent=False), this method calls solve_stationary().
        For time-dependent problems, the user must call solve_time(dt, T).

        Returns:
            np.ndarray: The computed solution vector (stationary) or time history (transient).
        """
        if self.time_dependent:
            raise NotImplementedError("For time-dependent problems, please use solve_time(dt, T).")
        else:
            return self.solve_stationary()