# pyfoam\pyfoam\core\solvers\fvm_heat_solver.py

import numpy as np
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

class HeatFVMSolver:
    def __init__(self,
                 mesh: StructuredMesh,
                 D: Callable[[np.ndarray], float],
                 S: Callable[[np.ndarray, float], float],
                 u0: Callable[[np.ndarray], float],
                 dirichlet_bc: Optional[DirichletCondition] = None,
                 neumann_bc: Optional[NeumannCondition] = None):
        """
        Initialize the FVM solver for the heat equation.

        Parameters:
            mesh (StructuredMesh): A structured mesh with attributes:
                                   - nodes: NumPy array of node coordinates.
                                   - cells: List of cells (each cell is a list of vertex indices).
                                   Also, the mesh should provide 'bounds' and 'num_cells'.
            D (Callable[[np.ndarray], float]): Diffusion coefficient function D(x,y).
            S (Callable[[np.ndarray, float], float]): Source term function S(x,y,t).
            u0 (Callable[[np.ndarray], float]): Initial condition function u₀(x,y) for cell averages.
            dirichlet_bc (Optional[DirichletCondition]): Dirichlet boundary condition object.
            neumann_bc (Optional[NeumannCondition]): Neumann boundary condition object.
        """
        self.mesh = mesh
        self.D = D
        self.S = S
        self.u0 = u0
        self.dirichlet_bc = dirichlet_bc
        self.neumann_bc = neumann_bc

        # Compute cell centers and volumes.
        self.cell_centers, self.cell_volumes, self.grid_shape = self._compute_cell_info()
        self.num_cells = self.cell_centers.shape[0]
        # Initialize cell-average solution using u0.
        self.u = np.array([self.u0(center) for center in self.cell_centers])
        # Will store the time history of the solution.
        self.solution_time = None

    def _compute_cell_info(self):
        """
        Compute cell centers and volumes for a structured rectangular mesh.

        For a 2D mesh with bounds [(x_min, x_max), (y_min, y_max)] and num_cells = [n_x, n_y]:
            dx = (x_max - x_min)/n_x, dy = (y_max - y_min)/n_y.
        The center of cell (i,j) is:
            (x_min + (i+0.5)*dx, y_min + (j+0.5)*dy),
        and the cell area is dx*dy.

        Returns:
            cell_centers (np.ndarray): Array of shape (n_cells, 2) with cell center coordinates.
            cell_volumes (np.ndarray): Array of cell areas.
            grid_shape (tuple): Tuple (n_x, n_y) representing the number of cells in each direction.
        """
        if not hasattr(self.mesh, 'num_cells'):
            raise ValueError("Mesh must have 'num_cells' attribute.")
        n_x, n_y = self.mesh.num_cells
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        dx = (x_max - x_min) / n_x
        dy = (y_max - y_min) / n_y

        centers = []
        volumes = []
        for i in range(n_x):
            for j in range(n_y):
                center = np.array([x_min + (i + 0.5) * dx, y_min + (j + 0.5) * dy])
                centers.append(center)
                volumes.append(dx * dy)
        return np.array(centers), np.array(volumes), (n_x, n_y)

    def _find_neighbors(self):
        """
        Build neighbor connectivity for cells in a structured 2D mesh.

        Cells are assumed to be arranged in grid_shape (n_x, n_y) and ordered in row-major order.
        Returns a dictionary mapping each cell index to its neighbor indices (with keys 'left', 'right',
        'up', and 'down'). If a neighbor does not exist (boundary), the value is None.

        Returns:
            neighbors (dict): Dictionary mapping cell indices to a dict of neighbor indices.
        """
        n_x, n_y = self.grid_shape
        neighbors = {}
        for i in range(n_x):
            for j in range(n_y):
                idx = i * n_y + j
                cell_nbr = {}
                cell_nbr['left']  = i * n_y + (j - 1) if j > 0 else None
                cell_nbr['right'] = i * n_y + (j + 1) if j < n_y - 1 else None
                cell_nbr['up']    = (i - 1) * n_y + j if i > 0 else None
                cell_nbr['down']  = (i + 1) * n_y + j if i < n_x - 1 else None
                neighbors[idx] = cell_nbr
        return neighbors

    def apply_boundary_conditions(self, u: np.ndarray, t: float) -> np.ndarray:
        """
        Apply Dirichlet boundary conditions to the cell-average solution.

        For each cell whose center lies on the domain boundary (within a tolerance),
        the DirichletCondition's function is used to prescribe the value.
        (Neumann conditions could also be added here to modify fluxes if desired.)

        Parameters:
            u (np.ndarray): Current cell-average solution (vector of length num_cells).
            t (float): Current time.

        Returns:
            np.ndarray: The solution vector with Dirichlet BCs enforced.
        """
        tol = 1e-8
        bounds = self.mesh.bounds
        for idx, center in enumerate(self.cell_centers):
            on_boundary = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(center[d], lower, atol=tol) or np.isclose(center[d], upper, atol=tol):
                    on_boundary = True
                    break
            if on_boundary and (self.dirichlet_bc is not None):
                u[idx] = self.dirichlet_bc.func(center, t)
        return u

    def solve(self, dt: float, T: float) -> np.ndarray:
        """
        Solve the transient heat equation using explicit Euler time stepping.

        The FVM formulation is based on conservation of energy over each control volume:
        
            V_i (du_i/dt) = Σ_faces F_face + V_i S(x_i,t),
        
        where u_i is the cell-average value, V_i is the cell volume, and F_face are the numerical fluxes.
        For interior cells in a uniform grid with cell size h (assuming dx = dy = h), the flux across an interior face
        between cells P and Q is approximated by:
        
            F = -0.5*(D(P)+D(Q))*(u_Q - u_P)/h * (face area).
        
        For a rectangular cell, the face areas are dy (for vertical faces) and dx (for horizontal faces).
        Dirichlet BCs override the update in boundary cells.
        
        Parameters:
            dt (float): Time step size.
            T (float): Final time.

        Returns:
            np.ndarray: A 2D array of shape (num_time_steps+1, num_cells) containing the cell-average solution at each time step.
        """
        num_steps = int(T / dt)
        neighbors = self._find_neighbors()
        u = self.u.copy()
        solution_time = np.zeros((num_steps + 1, self.num_cells))
        solution_time[0, :] = u.copy()

        # Extract grid spacing.
        n_x, n_y = self.grid_shape
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        dx = (x_max - x_min) / n_x
        dy = (y_max - y_min) / n_y

        for n in range(num_steps):
            t_next = (n + 1) * dt
            u_new = u.copy()
            # Loop over all cells.
            for idx in range(self.num_cells):
                # Check if cell is on the boundary.
                center = self.cell_centers[idx]
                is_boundary = False
                for d, (lower, upper) in enumerate(self.mesh.bounds):
                    if np.isclose(center[d], lower, atol=1e-8) or np.isclose(center[d], upper, atol=1e-8):
                        is_boundary = True
                        break
                # If boundary and Dirichlet BC exists, override.
                if is_boundary and (self.dirichlet_bc is not None):
                    u_new[idx] = self.dirichlet_bc.func(center, t_next)
                    continue

                # Otherwise, compute net flux from neighboring cells.
                net_flux = 0.0
                nbr = neighbors[idx]
                # Left face.
                if nbr['left'] is not None:
                    u_left = u[nbr['left']]
                    flux = -0.5*(self.D(center) + self.D(self.cell_centers[nbr['left']]))*(u[nbr['left']] - u[idx])/dx * dy
                    net_flux += flux
                # Right face.
                if nbr['right'] is not None:
                    u_right = u[nbr['right']]
                    flux = -0.5*(self.D(center) + self.D(self.cell_centers[nbr['right']]))*(u[nbr['right']] - u[idx])/dx * dy
                    net_flux += flux
                # Up face.
                if nbr['up'] is not None:
                    u_up = u[nbr['up']]
                    flux = -0.5*(self.D(center) + self.D(self.cell_centers[nbr['up']]))*(u[nbr['up']] - u[idx])/dy * dx
                    net_flux += flux
                # Down face.
                if nbr['down'] is not None:
                    u_down = u[nbr['down']]
                    flux = -0.5*(self.D(center) + self.D(self.cell_centers[nbr['down']]))*(u[nbr['down']] - u[idx])/dy * dx
                    net_flux += flux

                # Add source term contribution.
                S_val = self.S(center, t_next)
                # Update: u_new = u + dt/V*(net_flux + V S)
                u_new[idx] = u[idx] + dt/self.cell_volumes[idx]*(net_flux + self.cell_volumes[idx]*S_val)
            # Apply boundary conditions to u_new.
            u_new = self.apply_boundary_conditions(u_new, t_next)
            u = u_new.copy()
            solution_time[n+1, :] = u
        self.solution_time = solution_time
        return solution_time