# pyfoam\pyfoam\core\sovlers\fvm_poisson_solver.py

import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

class PoissonFVMSolver:
    def __init__(self,
                 mesh: StructuredMesh,
                 f: Callable[[np.ndarray], float],
                 dirichlet_bc: DirichletCondition,
                 neumann_bc: Optional[NeumannCondition] = None):
        """
        Initialize the Poisson FVM solver.

        Parameters:
            mesh (StructuredMesh): A structured mesh with attributes:
                                   - nodes: NumPy array of node coordinates.
                                   - cells: List of cells (each cell is a list of vertex indices).
                                   Additionally, the mesh should provide 'bounds' and 'num_cells'.
            f (Callable[[np.ndarray], float]): Source term function f(x,y).
            dirichlet_bc (DirichletCondition): Dirichlet BC object prescribing u = g(x,y) on ∂Ω.
            neumann_bc (Optional[NeumannCondition]): Optional Neumann BC object prescribing flux h(x,y) on ∂Ω.
        """
        self.mesh = mesh
        self.f = f
        self.dirichlet_bc = dirichlet_bc
        self.neumann_bc = neumann_bc

        # For FVM, the degrees of freedom are the cell-average values.
        # Compute cell centers and cell volumes.
        self.cell_centers, self.cell_volumes, self.grid_shape = self._compute_cell_info()
        self.num_cells = self.cell_centers.shape[0]
        self.solution = None  # Will hold the cell-average solution (flattened vector)

    def _compute_cell_info(self):
        """
        Compute cell centers and volumes for a structured rectangular mesh.

        For a 2D mesh with domain bounds [(x_min, x_max), (y_min, y_max)] and num_cells = [n_x, n_y]:
          - dx = (x_max - x_min)/n_x, dy = (y_max - y_min)/n_y.
          - The center of cell (i,j) is at:
                (x_min + (i+0.5)*dx, y_min + (j+0.5)*dy),
          - The cell area is dx*dy.
        
        Returns:
            cell_centers (np.ndarray): Array of shape (n_cells, 2) with cell center coordinates.
            cell_volumes (np.ndarray): Array of cell areas.
            grid_shape (tuple): (n_x, n_y).
        """
        if not hasattr(self.mesh, 'num_cells'):
            raise ValueError("The mesh object must have 'num_cells' attribute.")
        n_x, n_y = self.mesh.num_cells
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        dx = (x_max - x_min) / n_x
        dy = (y_max - y_min) / n_y

        centers = []
        volumes = []
        for i in range(n_x):
            for j in range(n_y):
                center = np.array([x_min + (i + 0.5) * dx, y_min + (j + 0.5) * dy])
                centers.append(center)
                volumes.append(dx * dy)
        return np.array(centers), np.array(volumes), (n_x, n_y)

    def _assemble_system(self) -> (sp.csr_matrix, np.ndarray):
        """
        Assemble the global linear system for the stationary Poisson equation
        using the finite volume method on a uniform rectangular grid.

        For interior control volumes, the discrete equation (five-point stencil) is:
            u_E + u_W + u_N + u_S - 4 u_P = - h^2 f_P,
        where h is the uniform cell size (assumed to be equal in x and y).

        Returns:
            A (sp.csr_matrix): Global coefficient matrix.
            b (np.ndarray): Global load vector.
        """
        # Retrieve grid parameters from mesh bounds and num_cells.
        n_x, n_y = self.grid_shape
        x_min, x_max = self.mesh.bounds[0]
        y_min, y_max = self.mesh.bounds[1]
        dx = (x_max - x_min) / n_x
        dy = (y_max - y_min) / n_y
        h = dx  # assuming dx = dy for a square cell

        N = self.num_cells  # Total number of cells.
        A_data = []
        A_rows = []
        A_cols = []
        b = np.zeros(N)

        # Helper: convert 2D cell index (i,j) to linear index.
        def lin_index(i, j):
            return i * n_y + j

        # Loop over cells.
        for i in range(n_x):
            for j in range(n_y):
                idx = lin_index(i, j)
                center = self.cell_centers[idx]
                f_val = self.f(center)  # Source term evaluated at cell center.
                # Discrete equation:
                # For interior cells:
                #   u_E + u_W + u_N + u_S - 4 u_P = - h^2 f(center)
                # For boundary cells, the equation will be overridden by Dirichlet conditions.
                # Set up the stencil if the cell is interior.
                if 0 < i < n_x - 1 and 0 < j < n_y - 1:
                    A_data.extend([1, 1, 1, 1, -4])
                    A_rows.extend([idx]*5)
                    A_cols.extend([
                        lin_index(i, j-1),  # left
                        lin_index(i, j+1),  # right
                        lin_index(i-1, j),  # up
                        lin_index(i+1, j),  # down
                        idx                # center
                    ])
                    b[idx] = - (h**2) * f_val
                else:
                    # For boundary cells, we set a placeholder; these will be replaced by BCs.
                    A_data.append(1.0)
                    A_rows.append(idx)
                    A_cols.append(idx)
                    b[idx] = - (h**2) * f_val

        A = sp.coo_matrix((A_data, (A_rows, A_cols)), shape=(N, N)).tocsr()
        return A, b

    def apply_boundary_conditions(self, A: sp.csr_matrix, b: np.ndarray) -> (sp.csr_matrix, np.ndarray):
        """
        Apply Dirichlet and Neumann boundary conditions to the system.

        For each cell whose center is on the boundary, the DirichletCondition's function is
        used to prescribe the value, i.e., the corresponding row of A is replaced with the identity
        and b is set to g(x,y). In addition, any Neumann flux contributions (if provided) are added to b.

        Parameters:
            A (sp.csr_matrix): Global coefficient matrix.
            b (np.ndarray): Global load vector.

        Returns:
            A, b: The modified matrix and vector with BCs enforced.
        """
        tol = 1e-8
        bounds = self.mesh.bounds
        for idx, center in enumerate(self.cell_centers):
            on_boundary = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(center[d], lower, atol=tol) or np.isclose(center[d], upper, atol=tol):
                    on_boundary = True
                    break
            if on_boundary:
                # Override with Dirichlet condition.
                value = self.dirichlet_bc.func(center, 0)  # t is not used here.
                A.data[A.indptr[idx]:A.indptr[idx+1]] = 0
                A[idx, idx] = 1.0
                b[idx] = value
        # Optionally, add Neumann contributions.
        if self.neumann_bc is not None:
            bc_flux = self.neumann_bc.apply(self.mesh, 0)
            for idx, flux in bc_flux.items():
                b[idx] += flux
        return A, b

    def solve(self) -> np.ndarray:
        """
        Solve the stationary Poisson equation using the finite volume method.

        The method assembles the global linear system, applies boundary conditions,
        and then solves A u = b for the cell-average values.

        Returns:
            np.ndarray: The computed cell-average solution vector.
        """
        A, b = self._assemble_system()
        A, b = self.apply_boundary_conditions(A, b)
        u = spla.spsolve(A, b)
        self.solution = u
        return u
