# pyfoam\pyfoam\core\solvers\fvm_solver.py

import numpy as np
from typing import Callable, Optional
from meshing.mesh import StructuredMesh
from conditions.condition import DirichletCondition, NeumannCondition

class FVMSolver:
    def __init__(self,
                 mesh: StructuredMesh,
                 D: Callable[[np.ndarray], float],
                 S: Callable[[np.ndarray, float], float],
                 u0: Callable[[np.ndarray], float],
                 dirichlet_bc: Optional[DirichletCondition] = None,
                 neumann_bc: Optional[NeumannCondition] = None):
        """
        Initialize the generic FVM solver.

        Parameters:
            mesh (StructuredMesh): A structured mesh with attributes:
                                   - nodes: NumPy array of node coordinates.
                                   - cells: List of cells (each cell is a list of vertex indices).
                                   Additionally, the mesh's domain bounds and number of cells are assumed.
            D (Callable[[np.ndarray], float]): Diffusion coefficient function D(x,y).
            S (Callable[[np.ndarray, float], float]): Source term function S(x,y,t).
            u0 (Callable[[np.ndarray], float]): Initial condition function u₀(x,y) for each control volume.
            dirichlet_bc (Optional[DirichletCondition]): (Optional) Dirichlet boundary condition object.
            neumann_bc (Optional[NeumannCondition]): (Optional) Neumann boundary condition object.
        """
        self.mesh = mesh
        self.D = D
        self.S = S
        self.u0 = u0
        self.dirichlet_bc = dirichlet_bc
        self.neumann_bc = neumann_bc

        # In FVM the degrees of freedom are the cell-averaged values.
        # Compute cell centers and cell volumes.
        self.cell_centers, self.cell_volumes, self.grid_shape = self._compute_cell_info()
        self.num_cells = len(self.cell_volumes)
        # Initialize the solution vector (one value per cell).
        self.u = np.array([self.u0(x) for x in self.cell_centers])
        # Time stepping parameters will be provided in the solve() method.
        self.solution_time = None

    def _compute_cell_info(self):
        """
        Compute cell centers and volumes for the structured mesh.

        For a structured rectangular mesh, each cell is defined by its vertices (from self.mesh.cells).
        The cell center is the average of the vertices' coordinates.
        The cell volume (area in 2D) is computed using the shoelace formula.
        Also, we return the grid shape (number of cells in each dimension) which can be deduced from the mesh.

        Returns:
            cell_centers (np.ndarray): Array of shape (num_cells, dim) with cell center coordinates.
            cell_volumes (np.ndarray): Array of length num_cells with cell volumes (areas).
            grid_shape (tuple): Tuple of cell counts in each dimension.
        """
        # We assume the StructuredMesh was created with num_cells list.
        # For convenience, we assume that the mesh object has an attribute "num_cells"
        # (if not, we can infer it from the mesh.cells length and the fact that cells are arranged in a grid).
        # Here we assume a 2D mesh.
        if not hasattr(self.mesh, 'num_cells'):
            raise ValueError("The mesh object must have a 'num_cells' attribute indicating the grid dimensions.")
        grid_shape = tuple(self.mesh.num_cells)
        cell_centers = []
        cell_volumes = []
        # For each cell in the mesh, compute the center and area.
        for cell in self.mesh.cells:
            vertices = self.mesh.nodes[cell]  # shape: (n_vertices, 2)
            center = np.mean(vertices, axis=0)
            cell_centers.append(center)
            # For a quadrilateral cell (structured mesh in 2D), assuming a rectangle,
            # the area can be computed by subdividing into two triangles.
            # Alternatively, if the mesh is uniform, area = dx*dy.
            # Here, we compute dx and dy from mesh bounds and num_cells.
            dx = (self.mesh.bounds[0][1] - self.mesh.bounds[0][0]) / self.mesh.num_cells[0]
            dy = (self.mesh.bounds[1][1] - self.mesh.bounds[1][0]) / self.mesh.num_cells[1]
            cell_volumes.append(dx * dy)
        return np.array(cell_centers), np.array(cell_volumes), grid_shape

    def _find_neighbors(self):
        """
        Build a neighbor connectivity for cells in the structured mesh.
        
        For a 2D structured mesh arranged in grid_shape (n_x, n_y), cells are ordered in row-major order.
        This method returns a dictionary mapping each cell index to its neighbor cell indices
        (with keys 'left', 'right', 'up', 'down'). If a neighbor does not exist (boundary), the value is None.
        
        Returns:
            neighbors (dict): A dictionary mapping cell indices (int) to a dict with keys 'left', 'right',
                              'up', 'down' and neighbor cell indices or None.
        """
        n_x, n_y = self.grid_shape
        neighbors = {}
        # Map cell index (i, j) in grid to linear index: idx = i*n_y + j.
        for i in range(n_x):
            for j in range(n_y):
                idx = i * n_y + j
                cell_neighbors = {}
                cell_neighbors['left'] = (i * n_y + (j - 1)) if j > 0 else None
                cell_neighbors['right'] = (i * n_y + (j + 1)) if j < n_y - 1 else None
                cell_neighbors['up'] = ((i - 1) * n_y + j) if i > 0 else None
                cell_neighbors['down'] = ((i + 1) * n_y + j) if i < n_x - 1 else None
                neighbors[idx] = cell_neighbors
        return neighbors

    def apply_boundary_conditions(self, u: np.ndarray, t: float) -> np.ndarray:
        """
        Apply Dirichlet boundary conditions to the cell-averaged solution.
        
        For a structured mesh, a cell is considered to be a boundary cell if its center lies
        outside the interior of the domain. This method uses the DirichletCondition object,
        which is assumed to provide g(x,t) for a given point.
        
        Parameters:
            u (np.ndarray): The current cell-average solution (one value per cell).
            t (float): Current time.
            
        Returns:
            np.ndarray: The solution vector with Dirichlet BCs applied.
        """
        # We assume that if the cell center is too close to the domain boundary, then a Dirichlet BC applies.
        bc_values = {}
        tol = 1e-8
        bounds = self.mesh.bounds  # e.g., [(x_min, x_max), (y_min, y_max)]
        for idx, center in enumerate(self.cell_centers):
            apply_bc = False
            for d, (lower, upper) in enumerate(bounds):
                if np.isclose(center[d], lower, atol=tol) or np.isclose(center[d], upper, atol=tol):
                    apply_bc = True
                    break
            if apply_bc:
                # Overwrite the cell value with the prescribed boundary value.
                # We use the DirichletCondition's function directly.
                bc_values[idx] = self.dirichlet_bc.func(center, t)
        # Apply the prescribed values.
        for idx, value in bc_values.items():
            u[idx] = value
        return u

    def solve(self, dt: float, T: float) -> np.ndarray:
        """
        Solve the time-dependent PDE using explicit Euler time-stepping.

        The FVM formulation is based on the integral conservation law over each control volume:
        
            V_i * (du_i/dt) = Σ_f F_f + V_i * S(x_i, t),
            
        where V_i is the volume of cell i, and F_f is the numerical flux across each face.
        
        For interior faces between cells i and j, the flux is approximated as:
        
            F_ij = - D_avg * (u_j - u_i) / d * A_f,
            
        where D_avg is the average diffusion coefficient evaluated at the centers of cells i and j,
        d is the distance between cell centers, and A_f is the face area.
        
        For boundary faces with Dirichlet BCs, the flux is computed using the prescribed boundary value.
        Neumann BC contributions are added directly.
        
        Parameters:
            dt (float): Time step size.
            T (float): Final time.
        
        Returns:
            np.ndarray: A 2D array of shape (num_time_steps+1, num_cells) containing the cell-average
                        solution at each time step.
        """
        num_steps = int(T / dt)
        # For simplicity, we assume a 2D structured mesh.
        # Compute cell centers and volumes already in __init__.
        neighbors = self._find_neighbors()
        u = self.u.copy()
        solution_time = np.zeros((num_steps + 1, self.num_cells))
        solution_time[0, :] = u.copy()

        # Extract grid spacings from mesh bounds and num_cells.
        dx = (self.mesh.bounds[0][1] - self.mesh.bounds[0][0]) / self.mesh.num_cells[0]
        dy = (self.mesh.bounds[1][1] - self.mesh.bounds[1][0]) / self.mesh.num_cells[1]

        # Face areas (for rectangular cells):
        area_vertical = dy    # left/right faces have area equal to dy.
        area_horizontal = dx  # top/bottom faces have area equal to dx.

        # Time-stepping loop.
        t = 0.0
        for n in range(num_steps):
            t_next = t + dt
            u_new = u.copy()
            # Loop over each cell.
            for idx in range(self.num_cells):
                # If the cell is on the boundary and a Dirichlet condition is specified, we override its value.
                center = self.cell_centers[idx]
                on_boundary = False
                for d, (lower, upper) in enumerate(self.mesh.bounds):
                    if np.isclose(center[d], lower, atol=1e-8) or np.isclose(center[d], upper, atol=1e-8):
                        on_boundary = True
                        break
                if on_boundary:
                    # Apply Dirichlet BC (overriding the update).
                    u_new[idx] = self.dirichlet_bc.func(center, t_next)
                    continue

                # Otherwise, compute net fluxes from neighboring cells.
                net_flux = 0.0
                nbr = neighbors[idx]
                # For each neighbor direction, compute flux.
                # Left neighbor.
                if nbr['left'] is not None:
                    d_center = self.cell_centers[nbr['left']]
                    flux = -0.5 * (self.D(center) + self.D(d_center)) * (u[nbr['left']] - u[idx]) / dx * area_vertical
                    net_flux += flux
                # Right neighbor.
                if nbr['right'] is not None:
                    d_center = self.cell_centers[nbr['right']]
                    flux = -0.5 * (self.D(center) + self.D(d_center)) * (u[nbr['right']] - u[idx]) / dx * area_vertical
                    net_flux += flux
                # Up neighbor.
                if nbr['up'] is not None:
                    d_center = self.cell_centers[nbr['up']]
                    flux = -0.5 * (self.D(center) + self.D(d_center)) * (u[nbr['up']] - u[idx]) / dy * area_horizontal
                    net_flux += flux
                # Down neighbor.
                if nbr['down'] is not None:
                    d_center = self.cell_centers[nbr['down']]
                    flux = -0.5 * (self.D(center) + self.D(d_center)) * (u[nbr['down']] - u[idx]) / dy * area_horizontal
                    net_flux += flux

                # Add source term contribution.
                S_val = self.S(center, t)
                # Update using explicit Euler:
                u_new[idx] = u[idx] + dt / self.cell_volumes[idx] * (net_flux + self.cell_volumes[idx] * S_val)
            # Update boundary cells via Dirichlet (Neumann BCs would modify net flux; here we assume they are incorporated if needed).
            u_new = self.apply_boundary_conditions(u_new, t_next)
            u = u_new.copy()
            solution_time[n + 1, :] = u
            t = t_next

        self.solution_time = solution_time
        return solution_time
